"""Configuration file handling for aib2ofx."""

import json
import os
import re


class Config:
    """Simple dictionary-like config object."""

    def __init__(self, config_filename='~/.aib2ofx.json'):
        filepath = open(os.path.expanduser(config_filename))
        config_string = filepath.read(-1)
        filepath.close()

        # Kill trailing commas.
        trailing_commas = re.compile(r',\s*([\]}])')
        config_string = trailing_commas.sub(r'\g<1>', config_string)
        self.cfg = json.loads(config_string)

    def get_config(self):
        """Returns the entire config object."""
        return self.cfg

    def users(self):
        """Returns list of configured users."""
        return list(self.cfg.keys())

    def __getitem__(self, name):
        if name in self.cfg:
            return self.cfg[name]
        raise AttributeError



"""All functionality related to interactions with AIB online interface."""

import csv
import datetime
import logging
import re
import tempfile
import time

import dateutil.parser as dparser
import mechanicalsoup


def _to_date(text):
    # AIB CSV export format: DD/MM/YYYY
    return dparser.parse(text, dayfirst=True, yearfirst=False)


def _to_value(text):
    tmp = text.strip().replace(',', '')
    if tmp[-3:] == ' DR':
        return '-' + tmp[:-3]
    return tmp


def _csv2account(csv_data, acc):
    transactions = list(csv_data)
    if not transactions:
        return None
    if 'Masked Card Number' in transactions[0]:
        acc['type'] = 'credit'
    else:
        acc['type'] = 'checking'
        acc['balance'] = transactions[-1].get('Balance', 0)
    operations = []
    for transaction in transactions:
        operation = {}
        operation['timestamp'] = _to_date(transaction['Posted Transactions Date'])
        # The mysterious story of 'Description' field in CSV exports continues!
        # Now the columns differ between CC and current account, on top of the
        # implemented bugs :(
        if acc['type'] == 'credit':
            desc = transaction['Description']
            if len(desc) > 18 and desc[18] == ' ':
                desc = desc[:18] + desc[19:]
        else:
            descriptions = []
            for i in [1, 2, 3]:
                descriptions.append(transaction['Description%s' % i].strip())
            desc = ' '.join(filter(bool, descriptions))
        operation['description'] = desc
        operation['debit'] = _to_value(transaction['Debit Amount'])
        operation['credit'] = _to_value(transaction['Credit Amount'])
        operations.append(operation)
    acc['operations'] = operations
    return acc


class CleansingFormatter(logging.Formatter):
    """Logging formatter that scrubs monetary values out."""

    def __init__(self, fmt=None, datefmt=None):
        self.amount_re = re.compile(r'(?:\d+,)*\d+\.\d+(?: DR)?')
        self.date_re = re.compile(r'\d\d/\d\d/\d\d')
        self.description_re = re.compile('<td>(?!dd/mm/yy).+</td>')
        logging.Formatter.__init__(self, fmt, datefmt)

    def format(self, record):
        tmp = record.msg
        tmp = self.amount_re.sub('X.XX', tmp)
        tmp = self.date_re.sub('dd/mm/yy', tmp)
        tmp = self.description_re.sub('<td>dummy description</td>', tmp)
        record.msg = tmp
        return logging.Formatter.format(self, record)


class Aib:
    """Automated browser interacting with AIB online interface."""

    def __init__(self, logindata, chatter):
        self.logindata = logindata
        if chatter['debug']:
            # make a directory for debugging output
            debugdir = tempfile.mkdtemp(prefix='aib2ofx_')
            print('WARNING: putting *sensitive* debug data in %s' % debugdir)
            self.logger = logging.getLogger("mechanize")
            logfile = logging.FileHandler(debugdir + '/mechanize.log', 'w')
            formatter = CleansingFormatter('%(asctime)s\n%(message)s')
            logfile.setFormatter(formatter)
            self.logger.addHandler(logfile)
            self.logger.setLevel(logging.DEBUG)
        else:
            logging.disable(logging.DEBUG)
            self.logger = logging.getLogger(None)
        self.browser = mechanicalsoup.StatefulBrowser()
        self.login_done = False
        self.data = {}

    def extract_value(self, varname):
        """Greps the page content for something that looks like a JS assignment, returns value."""
        response = str(self.browser.page)
        # Luckily the JS code isn't minified.
        regex = re.compile("%s = '([^']+)';" % varname)
        mangled_value = regex.search(response).group(1)
        value = (
            mangled_value.replace('\\/', '/')
            .replace('\\-', '-')
            .encode('latin1')
            .decode('unicode-escape')
        )
        # I feel dirty now. >_<
        return value

    def login(self):
        """Go through the login process."""
        brw = self.browser

        # Entry page.
        self.logger.debug('Requesting first login page.')
        brw.open('https://onlinebanking.aib.ie/inet/roi/login.htm')
        brw.select_form(selector='#loginCiamForm')
        self.logger.debug('Clicking large CONTINUE button on the entry page.')
        # Note: response code will be 401, as we haven't authorized yet.
        response = brw.submit_selected()
        assert response.status_code == 401

        # Redirect page.
        # This redirect is pure javascript, so we need to extract the target URL by hand.
        url = self.extract_value('window.location')
        # We're also saving one value that would be saved in session storage in a modern browser.
        encoded_post_params = self.extract_value('encodedPostParams')
        self.logger.debug('Bouncing through the interstitial.')
        response = brw.open(url, headers={'Referer': response.url})

        # Actual login form.
        brw.select_form()
        brw['pf.username'] = self.logindata['regNumber']
        brw['pf.pass'] = self.logindata['pin']
        self.logger.debug('Submitting login form.')
        brw.submit_selected()

        # Wait for 2FA on phone
        tfa_done = False
        while not tfa_done:
            brw.select_form('#finalizeForm')
            response = brw.submit_selected(update_state=False)
            if response.content == b'approved':
                tfa_done = True
            elif response.content != b'in_progress':
                raise RuntimeError(
                    'unexpected answer during 2FA auth: %s' % response.content
                )
            time.sleep(1)

        # Forward to normal interface.
        brw.select_form('#finalizeForm')
        response = brw.submit_selected()
        # This form is empty after page loads, fields are added by JS.
        form = brw.select_form(nr=0)
        form.new_control('hidden', 'state', self.extract_value('state.value'))
        form.new_control('hidden', 'nonce', self.extract_value('encodedNonce'))
        form.new_control('hidden', 'postParams', encoded_post_params)
        response = brw.submit_selected()
        assert response.status_code == 200
        brw.select_form(nr=0)
        response = brw.submit_selected()
        assert response.status_code == 200

        # mark login as done
        if brw.page.find(string='My Accounts'):
            self.login_done = True

    def get_data(self):
        """Download data for all accounts."""
        if not self.login_done:
            self.login()

        brw = self.browser
        self.data = {}
        # parse totals
        main_page = brw.page
        for account_line in main_page.find_all(
            'button', attrs={'class': 'account-button'}
        ):
            account_name = account_line.find('div', attrs={'class': 'account-name'})
            account_amount = account_line.find('span', attrs={'class': 'a-amount'})
            if not account_name:
                continue

            # Skip pension and saving accounts
            if account_line.find(text=re.compile('SAVINGS')):
                continue

            account = {}
            account['accountId'] = account_name.get_text(strip=True)
            account['available'] = _to_value(account_amount.get_text(strip=True))
            account['currency'] = 'EUR'
            account['bankId'] = 'AIB'
            account['reportDate'] = datetime.datetime.now()

            self.data[account['accountId']] = account

        # parse transactions
        #
        # Note: as of December 2021 there are *two* places that can produce CSV data:
        # 1. Accounts (top menu) > Transactions > Export (button)
        # 2. Accounts (top menu) > Transactions > Historical (button) > Export (button)
        #
        # They produce different set of CSV fields. Furthermore, second exporter
        # doesn't include any pending transactions, but offers wider date range.

        # To complicate things further: 'Historical Transactions' page available
        # upon pressing 'Historical' button used to be available directly from
        # the menu, but isn't anymore. We want to switch to that page *and then*
        # enumerate accounts as it saves time on going again through the 'Recent
        # Transactions' page. Given there's no direct link, we select the first
        # account, switch to historical transactions and iterate there. This
        # WILL fail if the first account doesn't have 'Historical' button :(
        self.logger.debug('Switching to historical transaction listing.')
        # Click 'Accounts > Transactions' in the top menu. This should bring up
        # first account's "Recent Transactions" page. This is usually the
        # current account, which has 'Historical' button.
        brw.select_form('#statement_form_id')
        brw.submit_selected()
        # Click 'Historical' button.
        brw.select_form('#historicalTransactionsCommand')
        brw.submit_selected()
        # We should be on the 'Historical Transactions' page.
        assert brw.page.find(string='Historical Transactions') != None

        # Interrogate the account dropdown.
        brw.select_form('#hForm')
        accounts_on_page = {
            o.text: o.get('value') for o in brw.form.form.find_all('option')
        }

        for account in list(self.data):
            if account not in accounts_on_page.keys():
                self.logger.debug(
                    'skipping account %s which is absent on historical'
                    'transactions page',
                    account,
                )
                del self.data[account]
                continue

            # get account's page
            self.logger.debug('Requesting transactions for %s.', account)
            brw.select_form('#hForm')
            brw['dsAccountIndex'] = accounts_on_page[account]
            brw.submit_selected()

            # click the export button
            form = brw.select_form('#historicalTransactionsCommand')
            # Some accounts (eg. freshly opened ones) have export facility
            # disabled. Skip them.
            if form.form.find(attrs={'name': 'export'}).get('value') == 'false':
                self.logger.debug(
                    'skipping account %s which has its "Export" button' 'disabled',
                    account,
                )
                del self.data[account]
                continue
            brw.submit_selected()

            # confirm the export request
            brw.select_form('#historicalTransactionsCommand')
            response = brw.submit_selected(update_state=False)
            csv_data = csv.DictReader(
                response.iter_lines(decode_unicode=True), skipinitialspace=True
            )
            self.data[account] = _csv2account(csv_data, self.data[account])

            # go back to the list of accounts
            brw.select_form('#historicaltransactions_form_id')
            brw.submit_selected()

    def getdata(self):
        """Returns data acquired from online interface."""
        return self.data

    def bye(self):
        """Logs user out of bank's online interface."""
        self.logger.debug('Logging out.')
        brw = self.browser
        brw.select_form('#formLogout')
        brw.submit_selected()
        if not brw.page.find(string='Logged Out'):
            raise Exception('Logout failed!')



"""Functionality related to producing OFX format out of parsed data."""

from hashlib import sha256
from xml.sax.saxutils import escape


_OPENING = """OFXHEADER:100
DATA:OFXSGML
VERSION:102
SECURITY:NONE
ENCODING:USASCII
CHARSET:1252
COMPRESSION:NONE
OLDFILEUID:NONE
NEWFILEUID:NONE

<OFX>
<SIGNONMSGSRSV1>
<SONRS>
<STATUS>
<CODE>0</CODE>
<SEVERITY>INFO</SEVERITY>
</STATUS><DTSERVER>%(reportDate)s</DTSERVER>
<LANGUAGE>ENG</LANGUAGE>
</SONRS>
</SIGNONMSGSRSV1>"""


_HEADERS = {
    'checking': """
<BANKMSGSRSV1>
<STMTTRNRS><TRNUID>1</TRNUID>
<STATUS><CODE>0</CODE><SEVERITY>INFO</SEVERITY></STATUS>
<STMTRS><CURDEF>%(currency)s</CURDEF>
<BANKACCTFROM><BANKID>%(bankId)s</BANKID>
<ACCTID>%(accountId)s</ACCTID>
<ACCTTYPE>CHECKING</ACCTTYPE>
</BANKACCTFROM>""",
    'credit': """
<CREDITCARDMSGSRSV1>
<CCSTMTTRNRS><TRNUID>1</TRNUID>
<STATUS><CODE>0</CODE><SEVERITY>INFO</SEVERITY></STATUS>
<CCSTMTRS><CURDEF>%(currency)s</CURDEF>
<CCACCTFROM>
<ACCTID>%(accountId)s</ACCTID>
</CCACCTFROM>""",
}


_TRANSACTIONS = """
<BANKTRANLIST>
<DTSTART>%(firstDate)s</DTSTART>
<DTEND>%(lastDate)s</DTEND>
%(transactions)s
</BANKTRANLIST>"""


_CLOSING = {
    'checking': """
<LEDGERBAL><BALAMT>%(balance)s</BALAMT><DTASOF>%(reportDate)s</DTASOF></LEDGERBAL>
<AVAILBAL><BALAMT>%(available)s</BALAMT><DTASOF>%(reportDate)s</DTASOF></AVAILBAL>
</STMTRS></STMTTRNRS></BANKMSGSRSV1></OFX>""",
    'credit': """
<LEDGERBAL><BALAMT>%(available)s</BALAMT><DTASOF>%(reportDate)s</DTASOF></LEDGERBAL>
</CCSTMTRS></CCSTMTTRNRS></CREDITCARDMSGSRSV1></OFX>""",
}


_SINGLE_TRANSACTION = """<STMTTRN>
<TRNTYPE>%(type)s</TRNTYPE>
<DTPOSTED>%(timestamp)s</DTPOSTED>
<TRNAMT>%(amount)s</TRNAMT>
<FITID>%(tid)s</FITID>
<NAME>%(description)s</NAME>
</STMTTRN>
"""


def _to_date(date):
    return date.strftime('%Y%m%d%H%M%S')


def bankdata_to_ofx(bank_data, later_than=None):
    """Turns dictionary into OFX document."""
    ofx = ''
    data = {}

    # Move obvious things to data.
    data = bank_data.copy()

    # Calculate rest of necessary fields.
    data['reportDate'] = _to_date(data['reportDate'])
    if data['operations']:
        data['firstDate'] = _to_date(data['operations'][0]['timestamp'])
        data['lastDate'] = _to_date(data['operations'][-1]['timestamp'])
    else:
        data['firstDate'] = data['reportDate']
        data['lastDate'] = data['reportDate']

    # Turn set of transactions into a lengthy string.
    hashes = {}
    transactions = []
    for transaction in data['operations']:
        trx = transaction.copy()
        if later_than and trx['timestamp'] <= later_than:
            continue
        trx['description'] = escape(trx['description'])
        if trx['credit'] and float(trx['credit']) != 0:
            trx['type'] = 'CREDIT'
            trx['amount'] = trx['credit']
        else:
            trx['type'] = 'DEBIT'
            trx['amount'] = '-%s' % trx['debit']
        trx['timestamp'] = _to_date(trx['timestamp'])
        hsh = sha256(
            trx['timestamp'].encode("utf-8")
            + trx['amount'].encode("utf-8")
            + trx['description'].encode("utf-8")
        )
        digest = hsh.hexdigest()
        # If there's been a transaction with identical hash in the current
        # set, record this and modify the hash to be different in OFX.
        if digest in hashes:
            count = hashes[digest] + 1
            hashes[digest] = count
            hsh.update(b'%d' % count)
        else:
            hashes[digest] = 1
        trx['tid'] = hsh.hexdigest()
        transactions.append(_SINGLE_TRANSACTION % trx)
    data['transactions'] = '\n'.join(transactions)

    # Wrap up and return resulting OFX.
    ofx = '\n'.join(
        (
            _OPENING,
            _HEADERS[data['type']],
            _TRANSACTIONS,
            _CLOSING[data['type']],
        )
    )
    ofx = ofx % data
    return ofx



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aib2ofx']

package_data = \
{'': ['*']}

install_requires = \
['mechanicalsoup>=1.3,<2.0', 'python-dateutil>=2.9.0,<3.0.0']

entry_points = \
{'console_scripts': ['aib2ofx = aib2ofx.cli:main']}

setup_kwargs = {
    'name': 'aib2ofx',
    'version': '0.73.2',
    'description': 'Download data from aib.ie in OFX format',
    'long_description': '# aib2ofx\n\n...or how to grab transaction data out of AIB\'s online interface, and format it\ninto `OFX` file.\n\n**NOTE:** Last AIB login update (Feb\' 2021) made me realise how brittle the\noverall machinery here is. The code that works around Web Storage API use is\nugly and likely to break. The most likely road forward for this project is to\ndecouple it into [ofxstatement](https://github.com/kedder/ofxstatement) plugin\nand (maybe) Selenium-powered CSV acquisition script. The former will be easy,\nthe latter will most likely be a nightmare to maintain and install, unless you\nenjoy having your banking details pipe through an arbitrary docker image.\n\nTime will tell.\n\n## Installation\n\n    python3 -mvenv aib2ofx\n    source aib2ofx/bin/activate\n    pip3 install aib2ofx\n\nThis will create a virtualenv for `aib2ofx`, fetch its code then install it with\nall dependencies. Once that completes, you\'ll find `aib2ofx` executable in the\n`bin` directory of this new virtualenv.\n\n## Usage\n\nCreate a `~/.aib2ofx.json` file, with AIB login details.\nSet the permission bits to 0600 to prevent other system users from reading it.\n\n    touch ~/.aib2ofx.json\n    chmod 0600 ~/.aib2ofx.json\n\nIt has a JSON format, single object with one key per AIB login you want to use.\n\n    {\n        "bradmajors": {\n            "regNumber": "12345678",\n            "pin": "12345"\n        }\n    }\n\nThe fields are as follows:\n\n* regNumber\n    > Your AIB registered number.\n\n* pin\n    > Your five digit PIN.\n\nYou can put more than one set of credentials in the file; the script\nwill download data for all accounts for all logins.\n\n    {\n        "bradmajors": {\n            "regNumber": "12345678",\n            "pin": "12345"\n        },\n        "janetweiss": {\n            "regNumber": "87654321",\n            "pin": "54321"\n        }\n    }\n\nNote that there\'s no comma after the last account details.\n\nOnce you\'ve prepared that config file, run:\n\n    aib2ofx -d /output/directory\n\nThe script should connect to AIB, log in using provided credentials,\niterate through all accounts, and save each of those to a separate\nfile located in `/output/directory`.\n\n## Guarantee\n\nThere is none.\n\nI\'ve written that script with my best intentions, it\'s not malicious,\nit\'s not sending the data anywhere, it\'s not doing anything nasty. I\'m\nusing it day to day to get data about my AIB accounts into a financial\nprogram that I use. It should work for you as good as it works for\nme. See the `LICENSE` file for more details.\n\n## Development\n\naib2ofx works only with python 3.\n\nIn order to set up a dev environment clone the repository, get\n[poetry](https://python-poetry.org/docs/#installation)\nand run `poetry install`. This will create a virtualenv with all\ndependencies installed. You can activate it with `poetry shell`.\n',
    'author': 'Jakub Turski',
    'author_email': 'yacoob@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/yacoob/aib2ofx',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.11',
}


setup(**setup_kwargs)



"""Command line interface of aib2ofx."""

import argparse
import datetime
import errno
import os
import re
import sys

import dateutil.parser as dparser

from . import aib
from . import cfg
from . import ofx


def get_options():
    """Parse argv into options."""
    parser = argparse.ArgumentParser(
        description='Download data from aib.ie in OFX format'
    )
    parser.add_argument(
        '-d',
        '--output-dir',
        default=os.path.join('.', datetime.date.today().strftime('%Y-%m-%d')),
        dest='output_dir',
        help='directory to put OFX files in [%(default)s]',
    )
    parser.add_argument(
        '-D',
        '--debug',
        action='store_true',
        default=False,
        dest='debug_mode',
        help='display some debug output [%(default)s]',
    )
    parser.add_argument(
        '-q',
        '--quiet',
        action='store_true',
        default=False,
        dest='quiet_mode',
        help='display no output at all [%(default)s]',
    )
    parser.add_argument(
        '-l',
        '--later-than',
        dest='later_than',
        help='exports only transactions later than specified date (YYYY-MM-DD)',
    )
    return parser.parse_args()


def write_file(output_dir, user, account_id, contents):
    """Save parsed data to a file."""
    outf = open('%s/%s_%s.ofx' % (output_dir, user, account_id), 'w')
    outf.write(contents)
    outf.close()


def get_data(user, config, output_dir, later_than, chatter):
    """Fetch, process and save data for a single user."""

    def show_and_tell(pre, function, post='done.'):
        if chatter['quiet']:
            function()
        else:
            print(pre, end=' ')
            sys.stdout.flush()
            function()
            print(post)

    cleanup_re = re.compile('[- 	]+')

    # Login to the bank, get data for all accounts.
    creds = config[user]
    bank = aib.Aib(creds, chatter)
    show_and_tell(
        'Logging in as \'%s\' (check your phone for 2FA)...' % user, bank.login
    )
    show_and_tell('Scraping account pages for data...', bank.get_data)
    show_and_tell('Logging \'%s\' out...' % user, bank.bye)

    # Save each account to separate OFX file.
    for account in bank.getdata().values():
        if not account:
            continue
        name = re.sub(cleanup_re, '_', account['accountId']).lower()
        contents = ofx.bankdata_to_ofx(account, later_than)
        write_file(output_dir, user, name, contents)


def main():
    """Main script entry point."""
    # Parse command line options.
    options = get_options()
    chatter = {
        'quiet': options.quiet_mode,
        'debug': options.debug_mode,
    }
    if options.later_than:
        later_than = dparser.parse(options.later_than, dayfirst=False, yearfirst=True)
    else:
        later_than = None

    # Read user-provided credentials.
    config = cfg.Config()

    try:
        os.makedirs(options.output_dir)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise

    # Iterate through accounts, scrape, format and save data.
    for user in config.users():
        get_data(user, config, options.output_dir, later_than, chatter)


if __name__ == '__main__':
    main()
